﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using Infragistics.Controls.Charts;
using System.Collections;
using System.Collections.Generic;
using XamWebDataChartSLApp.Model;
using System.Linq;

namespace XamWebDataChartSLApp.Behaviors
{
    public partial class ChartBehaviors : DependencyObject
    {
        #region HoveredCategoryItem Behavior
        public static readonly DependencyProperty HoveredCategoryItemsProperty =
           DependencyProperty.RegisterAttached("HoveredCategoryItems",
           typeof(HoveredCategoryItemsBehavior), typeof(ChartBehaviors),
           new PropertyMetadata(null,
               (o, e) => HoveredCategoryItemsChanged(
                   o as XamDataChart,
                   e.OldValue as HoveredCategoryItemsBehavior,
                   e.NewValue as HoveredCategoryItemsBehavior)));

        public static HoveredCategoryItemsBehavior GetHoveredCategoryItems(DependencyObject target)
        {
            return target.GetValue(HoveredCategoryItemsProperty)
                as HoveredCategoryItemsBehavior;
        }

        public static void SetHoveredCategoryItems(DependencyObject target, HoveredCategoryItemsBehavior behavior)
        {
            target.SetValue(HoveredCategoryItemsProperty, behavior);
        }

        private static void HoveredCategoryItemsChanged(
            XamDataChart chart,
            HoveredCategoryItemsBehavior oldValue,
            HoveredCategoryItemsBehavior newValue)
        {
            if (chart == null)
            {
                return;
            }
            if (oldValue != null)
            {
                oldValue.OnDetach(chart);
            }
            if (newValue != null)
            {
                newValue.OnAttach(chart);
            }
        }
        #endregion
    }

    public class HoveredCategoryItemsBehavior : FrameworkElement
    {
        public IComparer<object> ItemComparer { get; set; }
        public HoveredCategoryItemsBehavior()
        {
            HoveredItems = new SeriesItemInfoCollection();
        }

        private XamDataChart _owner = null;

        public static readonly DependencyProperty HoveredItemsProperty =
            DependencyProperty.Register("HoveredItems",
            typeof(SeriesItemInfoCollection),
            typeof(HoveredCategoryItemsBehavior),
            new PropertyMetadata(null));

        public SeriesItemInfoCollection HoveredItems
        {
            get { return (SeriesItemInfoCollection)GetValue(HoveredItemsProperty); }
            set { SetValue(HoveredItemsProperty, value); }
        }

        public bool CategoryOffset { get; set; }

        public void OnAttach(XamDataChart chart)
        {
            if (_owner != null)
            {
                OnDetach(_owner);
            }
            _owner = chart;
            chart.PropertyUpdated += Chart_PropertyUpdated;
            //((DataChart)chart.Parent).SeriesUpdated += Chart_SeriesUpdated;
        }

        public void OnDetach(XamDataChart chart)
        {
            if (_owner != chart)
            {
                return;
            }
            chart.PropertyUpdated -= Chart_PropertyUpdated;
            HoveredItems.Clear();
            _owner = null;
        }

        private void GetInViewAxisBounds(
            CategoryAxisBase x, NumericYAxis y,
            Rect viewportRect,
            out double left, out double right,
            out double top, out double bottom)
        {
            left = x.GetUnscaledValue(
                viewportRect.Left,
                _owner.WindowRect,
                viewportRect);
            right = x.GetUnscaledValue(
                viewportRect.Right,
                _owner.WindowRect,
                viewportRect);
            top = y.GetUnscaledValue(
                viewportRect.Top,
                _owner.WindowRect,
                viewportRect);
            bottom = y.GetUnscaledValue(
                viewportRect.Bottom,
                _owner.WindowRect,
                viewportRect);
        }

        private Rect GetViewportRect(
            Series target,
            CategoryAxisBase x,
            NumericYAxis y)
        {
            double top = y.TransformToVisual(target)
                    .Transform(new Point(0, 0)).Y;
            double bottom = y.TransformToVisual(target)
                .Transform(new Point(0, y.ActualHeight)).Y;
            double left = x.TransformToVisual(target)
                .Transform(new Point(0, 0)).X;
            double right = x.TransformToVisual(target)
                .Transform(new Point(x.ActualWidth, 0)).X;

            double width = right - left;
            double height = bottom - top;

            if (width > 0.0 && height > 0.0)
            {
                return new Rect(left, top, width, height);
            }
            return Rect.Empty;
        }

        private object GetItem(int index, Series series)
        {
            object item;

            try
            {
                if (series.ItemsSource is IList && index != int.MinValue && index < ((IList)series.ItemsSource).Count)
                {
                    item = ((IList)series.ItemsSource)[index];
                }
                else
                {
                    IEnumerator items = series.ItemsSource.GetType().GetGenericTypeDefinition() == typeof(ObservableQueue<>)
                    ? (series.ItemsSource as ObservableQueue<ChannelData>).GetReversedEnumerator()
                    : series.ItemsSource.GetEnumerator();
                    //IEnumerator items = series.ItemsSource.GetEnumerator();
                    for (int i = 0; i < index + 1; i++)
                    {
                        items.MoveNext();
                    }
                    item = items.Current;
                }
            }
            catch (Exception ex)
            {
                // Added a try/catch to help let the graph be visible and viewable when in design mode. Before this, it was crashing silverlight and throwing an exception box on mouse move
                item = null;
            }
            finally
            {
            }

            return item;
        }

        void Chart_SeriesUpdated()
        {
            // get the mouse point
            Point curPoint = new Point();
            UpdateHoveredItems(curPoint);
        }

        void Chart_PropertyUpdated(object sender, PropertyUpdatedEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine(e.PropertyName);
            if (e.PropertyName == "CrosshairPoint")
            {
                UpdateHoveredItems((Point)e.NewValue);
                //int seriesindex = -1;
                //foreach (CategorySeries series in _owner.Series.Where((s) => s is CategorySeries))
                //{
                //    //double start = DateTime.Now.Ticks;
                //    string type = "searched";
                //    if (seriesindex == -1)
                //    {
                //        CategoryAxisBase x = series.XAxis;
                //        NumericYAxis y = series.YAxis;
                //        Rect viewportRect = GetViewportRect(
                //           series,
                //           x,
                //           y);
                //        Point p = (Point)e.NewValue;

                //        double left;
                //        double right;
                //        double top;
                //        double bottom;
                //        GetInViewAxisBounds(x, y, viewportRect,
                //            out left, out right,
                //            out top, out bottom);
                //        double windowX = (p.X - _owner.WindowRect.Left) / _owner.WindowRect.Width;
                //        double xAxisValue = left + (windowX * (right - left));
                //        if (CategoryOffset)
                //        {
                //            xAxisValue -= .5;
                //        }
                //        xAxisValue = Math.Round(xAxisValue);
                //        object item = null;

                //        if (!double.IsInfinity(xAxisValue) &&
                //            !double.IsNaN(xAxisValue) &&
                //            xAxisValue >= 0 &&
                //            series.ItemsSource != null)
                //        {
                //            // This is for CategoryDateTimeXAxis
                //            // note, you certainly want to do something more  
                //            // efficient here, but its complexity depends on  
                //            // your data and whether it is pre-sorted, etc.  
                //            List<object> list = series.ItemsSource.OfType<object>().ToList();
                //            //list.Sort(ItemComparer);
                //            // This expects the list to be sorted from lowest to highest so if it is no then the index is from the other side
                //            int index = list.BinarySearch(new DateTime((long)xAxisValue), ItemComparer);
                //            if (index < 0)
                //            {
                //                index = (~index) - 1;
                //                //index = list.Count + index;
                //            }

                //            //this should actually determine if index or   
                //            //index + 1 is closer to the value we are searching for  
                //            //leaving this as excercise for the reader.  
                //            if (index > -1 && index < list.Count)
                //            {
                //                item = list[index];
                //            }

                //            if (item != null)
                //            {
                //                HoveredItems.UpdateSeriesItem(series, item);
                //            }

                //            seriesindex = index;
                //        }
                //    }
                //    else
                //    {
                //        type = "existing";
                //        // testing the assumption that since all series should have the exact same number of points
                //        // at all the same timestamps we should only have to find the index once
                //        object item = null;
                //        List<object> list = series.ItemsSource.OfType<object>().ToList();
                //        if (seriesindex > 0 && seriesindex < list.Count)
                //        {
                //            item = list[seriesindex];
                //        }

                //        if (item != null)
                //        {
                //            HoveredItems.UpdateSeriesItem(series, item);
                //        }
                //    }
                //    //System.Diagnostics.Debug.WriteLine(String.Format("Elapsed time ({0}): {1} ticks , used: {2}", series.Name, DateTime.Now.Ticks - start, type));
                //}
            }
        }

        void UpdateHoveredItems(Point point)
        {
            int seriesindex = -1;
            foreach (CategorySeries series in _owner.Series.Where((s) => s is CategorySeries))
            {
                //double start = DateTime.Now.Ticks;
                //string type = "searched";
                //if (seriesindex == -1)
                //{
                    CategoryAxisBase x = series.XAxis;
                    NumericYAxis y = series.YAxis;
                    Rect viewportRect = GetViewportRect(
                       series,
                       x,
                       y);
                    Point p = point;

                    double left;
                    double right;
                    double top;
                    double bottom;
                    GetInViewAxisBounds(x, y, viewportRect,
                        out left, out right,
                        out top, out bottom);
                    double windowX = (p.X - _owner.WindowRect.Left) / _owner.WindowRect.Width;
                    double xAxisValue = left + (windowX * (right - left));
                    if (CategoryOffset)
                    {
                        xAxisValue -= .5;
                    }
                    xAxisValue = Math.Round(xAxisValue);
                    object item = null;

                    if (!double.IsInfinity(xAxisValue) &&
                        !double.IsNaN(xAxisValue) &&
                        xAxisValue >= 0 &&
                        series.ItemsSource != null)
                    {
                        // This is for CategoryDateTimeXAxis
                        // note, you certainly want to do something more  
                        // efficient here, but its complexity depends on  
                        // your data and whether it is pre-sorted, etc.  
                        List<object> list = series.ItemsSource.OfType<object>().ToList();
                        //list.Sort(ItemComparer);
                        // This expects the list to be sorted from lowest to highest so if it is no then the index is from the other side
                        int index = list.BinarySearch(new DateTime((long)xAxisValue), ItemComparer);
                        if (index < 0)
                        {
                            index = (~index) - 1;
                            //index = list.Count + index;
                        }

                        //this should actually determine if index or   
                        //index + 1 is closer to the value we are searching for  
                        //leaving this as excercise for the reader.  
                        if (index > -1 && index < list.Count)
                        {
                            item = list[index];
                        }

                        if (item != null)
                        {
                            HoveredItems.UpdateSeriesItem(series, item);
                        }

                        seriesindex = index;
                    }
                //}
                //else
                //{
                //    type = "existing";
                //    // testing the assumption that since all series should have the exact same number of points
                //    // at all the same timestamps we should only have to find the index once
                //    object item = null;
                //    List<object> list = series.ItemsSource.OfType<object>().ToList();
                //    if (seriesindex > 0 && seriesindex < list.Count)
                //    {
                //        item = list[seriesindex];
                //    }

                //    if (item != null)
                //    {
                //        HoveredItems.UpdateSeriesItem(series, item);
                //    }
                //}
                //System.Diagnostics.Debug.WriteLine(String.Format("Elapsed time ({0}): {1} ticks , used: {2}", series.Name, DateTime.Now.Ticks - start, type));
            }
        }
    }
}
