﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using Infragistics.Controls.Charts;
using System.Windows.Browser;
using System.IO;
using System.Text;
using System.Collections.ObjectModel;
using System.Windows.Threading;
using System.ComponentModel;
using System.Windows.Data;
using System.Globalization;
using System.Windows.Media.Imaging;
using XamWebDataChartSLApp.Model;

namespace XamWebDataChartSLApp
{
    public partial class MainPage : UserControl
    {
        private ObservableCollection<ChannelDataGroup> historicalData = new ObservableCollection<ChannelDataGroup>();
        private const int queueCapacity = 3600;
        private Dictionary<int, ObservableQueue<ChannelData>> liveData = new Dictionary<int, ObservableQueue<ChannelData>>();
        private string serialNumber;
        private ObservableCollection<ChannelLookup> channelNames = new ObservableCollection<ChannelLookup>();
        private DateTime minhistoricaldatetime, maxhistoricaldatetime;
        private DateTime minlivedatetime, maxlivedatetime;

        public ChartViewModel Model { get; set; }
        public DataMode dataMode { get; set; }

        public MainPage()
        {
            InitializeComponent();
            dataMode = DataMode.Live;
            this.LayoutUpdated += new EventHandler(MainPage_LayoutUpdated);
            xmDataChart.PropertyUpdated += new PropertyUpdatedEventHandler(xmDataChart_PropertyUpdated);
            this.channelNames.CollectionChanged += new System.Collections.Specialized.NotifyCollectionChangedEventHandler(channelNames_CollectionChanged);
            Model = new ChartViewModel() { ChartZoomed = false };
            DataContext = Model;
        }

        void channelNames_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            // add/remove series and data for the specific channel
            if (e.Action == System.Collections.Specialized.NotifyCollectionChangedAction.Add && e.NewItems != null)
            {
                foreach (ChannelLookup channel in e.NewItems)
                {
                    // add the live data queue
                    AddLiveChannelQueue(channel.ChannelIndex);
                    AddHistoricalChannelData(channel.ChannelIndex);
                }
            }
            else if (e.OldItems != null)
            {
                // remove series
                foreach (ChannelLookup channel in e.OldItems)
                {
                    // remove data from live queue and then refresh the graph since this is not an observable collection
                    liveData.Remove(channel.ChannelIndex);
                    RemoveLayer(channel.ChannelName);

                    // just delete the historical data and the graph will automatically update
                    // Have to use linq since Exists() and FirstOrDefault() is not available in SL and 
                    // First() blows up if it is not found
                    var group = (from curr in historicalData
                                    where curr.index == channel.ChannelIndex
                                    select curr);

                    if (group != null && group.Count() > 0)
                        historicalData.Remove(group.First());
                }
            }
        }
        private void AddSeries(int channelIndex)
        {
            ChannelLookup lookup = channelNames.First(c => c.ChannelIndex == channelIndex);

            // create new Y axis
            NumericYAxis yaxis = new NumericYAxis();
            AxisLabelSettings labelsettings = new AxisLabelSettings();
            labelsettings.Visibility = System.Windows.Visibility.Collapsed;
            labelsettings.Location = AxisLabelsLocation.OutsideLeft;
            yaxis.LabelSettings = labelsettings;
            yaxis.Label = null;
            yaxis.Style = Resources["NumericYAxisStyle"] as Style;
            yaxis.MinimumValue = lookup.Min;
            yaxis.MaximumValue = lookup.Max;
            yaxis.Name = String.Format("y_{0}", lookup.ChannelName);
            xmDataChart.Axes.Add(yaxis);

            AxisLabelSettings xlabelsettings = new AxisLabelSettings();
            xlabelsettings.Visibility = System.Windows.Visibility.Collapsed;
            xlabelsettings.Location = AxisLabelsLocation.InsideBottom;
            CategoryDateTimeXAxis xaxis = new CategoryDateTimeXAxis();
            xaxis.LabelSettings = xlabelsettings;
            xaxis.Label = null;
            xaxis.Style = Resources["CategoryXAxisStyle"] as Style;
            xaxis.DateTimeMemberPath = "TimeStamp";
            xaxis.Name = String.Format("x_{0}", lookup.ChannelName);
            xaxis.ItemsSource = liveData[channelIndex];
            xmDataChart.Axes.Add(xaxis);

            DoubleCollection dashArray = new DoubleCollection();
            dashArray.Add(3.0);
            dashArray.Add(1.0);

            // Create a new series
            StepLineSeries series = new StepLineSeries();
            series.Title = lookup.ChannelName;
            series.Name = String.Format("series_{0}", lookup.ChannelName);
            series.MarkerType = MarkerType.None;
            series.ValueMemberPath = "Value";
            series.XAxis = xaxis;
            //if (xaxis.ItemsSource == null) { xaxis.ItemsSource = liveData[channelIndex]; }
            //series.XAxis.ItemsSource = liveData[channelIndex];
            series.YAxis = yaxis;
            //series.DashArray = dashArray;
            series.ItemsSource = liveData[channelIndex];
            xmDataChart.Series.Add(series);

            // Set the scatter line range
            DateTime maxTimeStamp = liveData[channelIndex].Max(cv => cv.TimeStamp);
            DateTime minTimeStamp = liveData[channelIndex].Min(cv => cv.TimeStamp);
            maxlivedatetime = maxlivedatetime < maxTimeStamp ? maxTimeStamp : maxlivedatetime;
            minlivedatetime = minlivedatetime > maxlivedatetime.AddSeconds(queueCapacity * (-1)) ? minlivedatetime : maxlivedatetime.AddSeconds(queueCapacity * (-1));
            SetScatterLineRange();
        }
        private void AddHistoricalSeries(int channelIndex)
        {
            ChannelLookup lookup = channelNames.First(c => c.ChannelIndex == channelIndex);
            List<ChannelData> data = historicalData.First(d => d.index == channelIndex).ChannelDataList;

            if (data.Count > 0)
            {
                // get the min date and max date so that all the channels will line up properly
                minhistoricaldatetime = minhistoricaldatetime != DateTime.MinValue && minhistoricaldatetime < data.Min(d => d.TimeStamp)
                    ? minhistoricaldatetime
                    : data.Min(d => d.TimeStamp);
                maxhistoricaldatetime = maxhistoricaldatetime > data.Max(d => d.TimeStamp)
                    ? maxhistoricaldatetime
                    : data.Max(d => d.TimeStamp);
            }

            // create new Y axis
            NumericYAxis yaxis = new NumericYAxis();
            AxisLabelSettings labelsettings = new AxisLabelSettings();
            labelsettings.Visibility = System.Windows.Visibility.Collapsed;
            labelsettings.Location = AxisLabelsLocation.InsideLeft;
            yaxis.LabelSettings = labelsettings;
            yaxis.Style = Resources["NumericYAxisStyle"] as Style;
            yaxis.MinimumValue = lookup.Min;
            yaxis.MaximumValue = lookup.Max;
            yaxis.Name = String.Format("y_{0}", lookup.ChannelName);
            xmDataChart.Axes.Add(yaxis);

            AxisLabelSettings xlabelsettings = new AxisLabelSettings();
            xlabelsettings.Visibility = System.Windows.Visibility.Collapsed;
            xlabelsettings.Location = AxisLabelsLocation.InsideBottom;
            CategoryDateTimeXAxis xaxis = new CategoryDateTimeXAxis();
            xaxis.LabelSettings = xlabelsettings;
            xaxis.Style = Resources["CategoryXAxisStyle"] as Style;
            xaxis.DateTimeMemberPath = "TimeStamp";
            xaxis.Name = String.Format("x_{0}", lookup.ChannelName);
            xaxis.ItemsSource = data.Count == 0 ? null : data; // binding a empty list keeps to the xaxis causing a "value out of range" error so set it to null instead
            xmDataChart.Axes.Add(xaxis);

            // create a stepline series
            StepLineSeries series = new StepLineSeries();
            series.Title = lookup.ChannelName;
            series.Name = String.Format("series_{0}", lookup.ChannelName);
            series.MarkerType = MarkerType.None;
            series.ValueMemberPath = "Value";
            series.XAxis = xaxis;
            //series.XAxis.ItemsSource = data;
            series.YAxis = yaxis;
            series.ItemsSource = data;
            xmDataChart.Series.Add(series);

            // Set the scatter line range only if data is returned for the channel.
            // All channels that have data should have the same number of readings
            // so we can always set the range if there is data.
            if (data != null && data.Count > 0)
                SetScatterLineRange();
        }
        private void LoadData()
        {
            if (channelNames.Count > 0)
                LoadData(channelNames.Select(cn => cn.ChannelIndex).ToList());
        }
        private void LoadData(List<int> indexes)
        {
            busyIndicator.IsBusy = true;
            busyIndicator.BusyContent = "Loading Chart Data...";
            List<ChannelData> sampledata = GenerateSampleData(indexes);
            foreach (var channelindex in sampledata.Select(c => c.ChannelIndex).Distinct())
            {
                historicalData.Add(new ChannelDataGroup() { index = channelindex, ChannelDataList = sampledata.Where(c => c.ChannelIndex == channelindex).OrderBy(c => c.TimeStamp).ToList() });
                AddHistoricalSeries(channelindex);
            }
            busyIndicator.IsBusy = false;
        }
        private void AddToQueue(ObservableQueue<ChannelData> queue, ChannelData channelData)
        {
            if (queue.Count >= queueCapacity)
                queue.Dequeue();
            queue.Enqueue(channelData);
        }
        private void AddLiveChannelQueue(int channelIndex)
        {
            if (!liveData.ContainsKey(channelIndex))
            {
                busyIndicator.IsBusy = true;
                busyIndicator.BusyContent = "Loading Chart Data...";
                // create the queue
                ObservableQueue<ChannelData> tempqueue = new ObservableQueue<ChannelData>(queueCapacity);
                tempqueue.Enqueue(new ChannelData() { TimeStamp = DateTime.Now, Value = null });

                // pre pop the collection with nulls
                DateTime past = DateTime.Now;
                for (int x = 0; x < queueCapacity; x++)
                {
                    tempqueue.Enqueue(new ChannelData() { TimeStamp = past.Subtract(new TimeSpan(0, 0, queueCapacity - x)), Value = null });
                }

                liveData.Add(channelIndex, tempqueue);

                if (dataMode == DataMode.Live)
                {
                    // Create the series mapping
                    AddSeries(channelIndex);
                }
				busyIndicator.IsBusy = false;
            }
        }
        private void AddHistoricalChannelData(int index)
        {
            if (dataMode == DataMode.Historical)
            {
                LoadData(new List<int>() { index });
            }
        }
        private void RemoveLayer(string name)
        {
            xmDataChart.Series.Remove(xmDataChart.Series[String.Format("series_{0}", name)]);
            xmDataChart.Axes.Remove(xmDataChart.Axes[String.Format("x_{0}", name)]);
            xmDataChart.Axes.Remove(xmDataChart.Axes[String.Format("y_{0}", name)]);
        }
        private void UpdateQueueData(Viper viperunit)
        {
            if (viperunit == null)
                return;

            AddQueueData(viperunit);
        }
        private void AddQueueData(Viper viperunit)
        {
            // Loop through each channel data item and add it to its respective queue but 
            // only if that index is in the channelnames. That list contains all the names/indexes
            // that the graph should be plotting.

            DateTime lastTimeStamp = DateTime.MinValue;
            foreach (var reading in viperunit.ChannelDataList.Where(d => channelNames.Count(cn => cn.ChannelIndex == d.ChannelIndex) > 0).OrderBy(d => d.TimeStamp))
            {
                // Create the queue and add the layer to the graph if it is not already there
                AddLiveChannelQueue(reading.ChannelIndex);

                // Add new item to the queue
                ObservableQueue<ChannelData> queue = liveData[reading.ChannelIndex];

                // The queue will contain data for every second so if there is more than a 1 second
                // gap between this new data and the last data entry in the queue then create new
                // entries with the last value up to, but not including the new data timestamp.
                //while (queue.Last().TimeStamp < reading.TimeStamp.AddSeconds(-1))
                //{
                //    AddToQueue(queue,
                //        new ChannelData()
                //        {
                //            ChannelIndex = queue.Last().ChannelIndex,
                //            TimeStamp = queue.Last().TimeStamp.AddSeconds(1),
                //            Value = queue.Last().Value
                //        }
                //    );
                //}
                if (reading.TimeStamp > queue.Last().TimeStamp) // this should always be the case
                {
                    maxlivedatetime = maxlivedatetime < reading.TimeStamp ? reading.TimeStamp : maxlivedatetime;
                    minlivedatetime = minlivedatetime > maxlivedatetime.AddSeconds(queueCapacity * (-1)) ? minlivedatetime : maxlivedatetime.AddSeconds(queueCapacity * (-1));
                    // Add new entry
                    AddToQueue(queue, reading);
                }

                lastTimeStamp = reading.TimeStamp;
            }

            // all readings for this dataset have been added so make sure all queues 
            // have entries up to and including the lastTimeStamp
            //foreach (var q in liveData.Values)
            //{
            //    while (q.Last().TimeStamp < lastTimeStamp)
            //    {
            //        AddToQueue(q,
            //            new ChannelData()
            //            {
            //                ChannelIndex = q.Last().ChannelIndex,
            //                TimeStamp = q.Last().TimeStamp.AddSeconds(1),
            //                Value = q.Last().Value
            //            }
            //        );
            //    }
            //}

            // Update the scatter line range and interval based on the first queue 
            // (all queues are supposed to have the same number of points and first/last timestamps)
            if (liveData.Count > 0 && dataMode == DataMode.Live)
                SetScatterLineRange();
        }
        private void SetHorizontalZoom(bool enable)
        {
            xmDataChart.WindowScaleHorizontal = 1;
            if (enable)
            {
                xmDataChart.HorizontalZoombarVisibility = System.Windows.Visibility.Visible;
                xmDataChart.HorizontalZoomable = true;
            }
            else
            {
                xmDataChart.HorizontalZoombarVisibility = System.Windows.Visibility.Collapsed;
                xmDataChart.HorizontalZoomable = false;
            }
        }
        private void ToggleChartView()
        {
            foreach (var series in xmDataChart.Series.OfType<StepLineSeries>())
            {
                // Remove the x and y axes for each series
                xmDataChart.Axes.Remove(xmDataChart.Axes[String.Format("x_{0}", series.Name.ToString().Remove(0, 7))]);
                xmDataChart.Axes.Remove(xmDataChart.Axes[String.Format("y_{0}", series.Name.ToString().Remove(0, 7))]);
            }
            xmDataChart.Series.Clear();


            if (dataMode == DataMode.Historical) // set to historical mode
            {
                // Make sure the horizontal zoom bar is visible
                SetHorizontalZoom(true);
                LoadData();
            }
            else // set to Live mode
            {
                // turn off the horizontal zoom bar when in live view
                SetHorizontalZoom(false);
                CategoryDateTimeXAxis axis = xmDataChart.Axes["XAxis"] as CategoryDateTimeXAxis;
                foreach (var channel in liveData)
                {
                    AddSeries(channel.Key);
                }
            }
        }

        #region Get Crosshair values
        private void xmDataChart_PropertyUpdated(object sender, Infragistics.Controls.Charts.PropertyUpdatedEventArgs e)
        {
            switch (e.PropertyName)
            {
                case "WindowRect":
                    if (e.NewValue != null && e.NewValue is Rect)
                    {
                        Rect oldview = (Rect)e.OldValue;
                        Rect newview = (Rect)e.NewValue;
                        if (oldview.Width != newview.Width)
                        {
                            // we zoomed so reset the x axis interval
                            SetScatterLineRange();
                            Model.ChartZoomed = newview.Width < 1;
                        }
                    }
                    break;
            }
        }
        #endregion

        #region User events
        private void btnHistory_Click(object sender, RoutedEventArgs e)
        {
            // toggle the datamode
            dataMode = dataMode == DataMode.Live ? DataMode.Historical : DataMode.Live;

            string path = dataMode == DataMode.Live ? "images/clock_play.png" : "images/clock_stop.png";
            Button button = sender as Button;
            Image image = button.Content as Image;
            image.Source = new BitmapImage(new Uri(path, UriKind.Relative));
            ToolTipService.SetToolTip(button, dataMode == DataMode.Live ? "Show History" : "Show Live Data");
            ToggleChartView();
        }
        private void RadioButton_Click(object sender, RoutedEventArgs e)
        {
            if (serialNumber != "-1")
            {
                // toggle the datamode
                dataMode = ((RadioButton)sender).Name == "radioHistory" ? DataMode.Historical : DataMode.Live;
                ToggleChartView();
            }
        }
        private void btnZoom_Click(object sender, RoutedEventArgs e)
        {
            xmDataChart.WindowScaleHorizontal = 1;
        }
        #endregion

        #region Sample Data generation
        DispatcherTimer t = new DispatcherTimer();

        private void UserControl_Loaded(object sender, RoutedEventArgs e)
        {
            // Timer is strictly for sample "live" data
            t.Interval = new TimeSpan(0, 0, 0, 0, 500);
            t.Tick += new EventHandler(UpdateDataDelegate);
            t.Start();
            //SetupChart();
            //LoadData();
        }
        private void UpdateDataDelegate(object sender, EventArgs e)
        {
            Random r = new Random();
            List<ChannelData> channeldata = new List<ChannelData>();
            //channeldata.Add(new ChannelData() { ChannelIndex = 1, Value = 50 + (10 * Math.Sin(DateTime.Now.Second * 6)), TimeStamp = DateTime.Now });
            //channeldata.Add(new ChannelData() { ChannelIndex = 2, Value = 30 + (10 * Math.Sin(DateTime.Now.Second * 6)), TimeStamp = DateTime.Now });
            //channeldata.Add(new ChannelData() { ChannelIndex = 3, Value = 10 + (10 * Math.Sin(DateTime.Now.Second * 6)), TimeStamp = DateTime.Now });
            //channeldata.Add(new ChannelData() { ChannelIndex = 4, Value = 70 + (10 * Math.Sin(DateTime.Now.Second * 6)), TimeStamp = DateTime.Now });
            channeldata.Add(new ChannelData() { ChannelIndex = 1, Value = 50 + r.Next(-8, 8), TimeStamp = DateTime.Now });
            channeldata.Add(new ChannelData() { ChannelIndex = 2, Value = 30 + r.Next(-8, 8), TimeStamp = DateTime.Now });
            channeldata.Add(new ChannelData() { ChannelIndex = 3, Value = 10 + r.Next(-8, 8), TimeStamp = DateTime.Now });
            channeldata.Add(new ChannelData() { ChannelIndex = 4, Value = 70 + r.Next(-8, 8), TimeStamp = DateTime.Now });

            Viper viperunit = new Viper()
            {
                SerialNumber = "1234",
                ViperId = 1,
                Status = 1,
                ChannelDataList = channeldata
            };

            UpdateQueueData(viperunit);
        }
        private List<ChannelData> GenerateSampleData(List<int> indexes)
        {
            Random r = new Random();

            DateTime time = DateTime.Now;
            List<ChannelData> dataList = new List<ChannelData>();
            foreach (int index in indexes)
            {
                int range = ((100 / indexes.Count) / 2) * index; // separate the series a bit
                for (int x = 0; x < 2500 * index; x++)
                {
                    dataList.Add(new ChannelData() { ChannelIndex = index, Value = r.Next(range - 4, range + 4), TimeStamp = time.AddSeconds(x) });
                }
            }
            return dataList;
        }
        void MainPage_LayoutUpdated(object sender, EventArgs e)
        {
            // Put the chart into live mode
            // add the four names into the names list
            // start the timer
            if (channelNames.Count < 4)
            {
                dataMode = DataMode.Live;
                channelNames.Add(new ChannelLookup() { ChannelIndex = 1, ChannelName = "SvO2", Max = 100, Min = 0, Units = "%" });
                channelNames.Add(new ChannelLookup() { ChannelIndex = 2, ChannelName = "HCT", Max = 100, Min = 0, Units = "%" });
                //channelNames.Add(new ChannelLookup() { ChannelIndex = 3, ChannelName = "Flow", Max = 100, Min = 0, Units = "l/min" });
                //channelNames.Add(new ChannelLookup() { ChannelIndex = 4, ChannelName = "Temp", Max = 100, Min = 0, Units = "f" });
                SetHorizontalZoom(false);
            }
        }
        #endregion

        #region Helper functions
        private int DetermineXAxisInterval(long startTick, long endTick, double width)
        {
            if (startTick == 0 || endTick == 0 || startTick == DateTime.MinValue.Ticks || endTick == DateTime.MaxValue.Ticks)
                return -1;

            // determine the interval
            int interval = (int)Math.Abs(Math.Ceiling((((endTick - startTick) / TimeSpan.TicksPerSecond) * width) / 5)); // we want 5 time labels

            if (interval < 2)
            {
                interval = 1;
            }
            else if (interval <= 120)
            {
                switch ((int)Math.Floor(interval / 10))
                {
                    case 0:
                        interval = 5;
                        break;
                    case 1:
                        interval = 15;
                        break;
                    case 2:
                    case 3:
                        interval = 30;
                        break;
                    case 4:
                    case 5:
                    case 6:
                        interval = 60;
                        break;
                    default:
                        interval = 120;
                        break;
                }
            }
            else
            {
                interval = ((int)Math.Floor(interval / 300) + 1) * 300;
            }

            return interval;
        }
        private void SetScatterLineRange()
        {
            DateTime mindate, maxdate;

            if (dataMode == DataMode.Live)
            {
                mindate = minlivedatetime;
                maxdate = maxlivedatetime;
            }
            else
            {
                mindate = minhistoricaldatetime;
                maxdate = maxhistoricaldatetime;
            }

            // set the range for all series
            foreach (var xaxis in xmDataChart.Axes.OfType<CategoryDateTimeXAxis>())
            {
                xaxis.MinimumValue = mindate;
                xaxis.MaximumValue = maxdate;
            }

            NumericXAxis a = xmDataChart.Axes.OfType<NumericXAxis>().First();
            a.MinimumValue = mindate.Ticks;
            a.MaximumValue = maxdate.Ticks;

            a.Interval = TimeSpan.FromSeconds(DetermineXAxisInterval((long)a.MinimumValue, (long)a.MaximumValue, xmDataChart.WindowRect.Width)).Ticks;
        }
        #endregion
    }
}