﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;

namespace XamWebDataChartSLApp.Model
{
    public class ObservableQueue<T> : IEnumerable<T>, INotifyCollectionChanged
    {
        Queue<T> _internalqueue;
        int _capacity;

        public ObservableQueue(int Capacity)
        {
            _capacity = Capacity;
            _internalqueue = new Queue<T>(_capacity);
        }

        public void Enqueue(T obj)
        {
            _internalqueue.Enqueue(obj);
            OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, obj, this.Count - 1));
        }

        public T Dequeue()
        {
            T obj = _internalqueue.Dequeue();
            OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, obj, 0));
            return obj;
        }

        public int Count
        {
            get { return _internalqueue.Count; }
        }

        #region INotifyCollectionChanged Members

        public event NotifyCollectionChangedEventHandler CollectionChanged;

        protected virtual void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
        {
            if (CollectionChanged != null)
            {
                CollectionChanged(this, e);
            }
        }
        #endregion

        #region IEnumerable<T> Members

        public IEnumerator<T> GetEnumerator()
        {
            return _internalqueue.GetEnumerator();
        }
        public IEnumerator<T> GetReversedEnumerator()
        {
            return _internalqueue.Reverse().GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return _internalqueue.GetEnumerator();
        }

        #endregion
    }
}
